//
// KeyboardWindow.cpp
//
// Copyright (c) 2003, Roku, LLC.  All rights reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "KeyboardWindow.h"

///////////////////////////////////////////////////////////////////////////////
// module KeyboardWindow #defines

#define TRANSPARENCY 127
#define _3D_RECT_COLOR_LIGHT             CascadeColor(164, 164, 164, TRANSPARENCY)
#define _3D_RECT_COLOR_DARK              CascadeColor(52, 52, 52, TRANSPARENCY)

///////////////////////////////////////////////////////////////////////////////
// module KeyboardWindow static functions

static void Draw3DRect(CascadeBitmap & bitmap, const CascadeRect & rect, bool bRaised = true)
{
	const CascadeColor colorTopLeft  = bRaised ? _3D_RECT_COLOR_LIGHT : _3D_RECT_COLOR_DARK;
	const CascadeColor colorBotRight = bRaised ? _3D_RECT_COLOR_DARK : _3D_RECT_COLOR_LIGHT;

	CascadeRect r = rect;
    bitmap.DrawLine(r.TopLeft(), r.TopRight(), colorTopLeft);
    bitmap.DrawLine(r.TopLeft(), r.BottomLeft(), colorTopLeft);
    bitmap.DrawLine(r.TopRight(), r.BottomRight(), colorBotRight);
    bitmap.DrawLine(r.BottomLeft(), r.BottomRight(), colorBotRight);
    r.InflateRect(-1, -1);
    bitmap.DrawLine(r.TopLeft(), r.TopRight(), colorTopLeft);
    bitmap.DrawLine(r.TopLeft(), r.BottomLeft(), colorTopLeft);
    bitmap.DrawLine(r.TopRight(), r.BottomRight(), colorBotRight);
    bitmap.DrawLine(r.BottomLeft(), r.BottomRight(), colorBotRight);
}

#if 0
static void DrawSunkenRect(CascadeBitmap & bitmap, const CascadeRect & rect)
{
	Draw3DRect(bitmap, rect, false);
}
#endif

///////////////////////////////////////////////////////////////////////////////
// class KeyboardWindow::LilTextWindow

void
KeyboardWindow::LilTextWindow::OnPaint(CascadeBitmap & bitmap)
{
	//printf("repaint letter %s active %d\n", (const char *) m_text, IsActiveFocus());
	KeyboardWindow * pKeyboardWindow = (KeyboardWindow *) GetParent();

	// paint background
	{
		CascadeRect clipRect = bitmap.GetClipRect();
		bitmap.SetClipRect(GetRectAbsolute());
		pKeyboardWindow->PaintBackground(bitmap);
		bitmap.SetClipRect(clipRect);
	}

	u32 flags = TEXT_CENTER_VERTICALLY | TEXT_CENTER_HORIZONTALLY;
	CascadeRect rect = GetRectAbsolute();
	if (m_text.GetLength() > 1)
	{
		flags = TEXT_CENTER_VERTICALLY;
		rect.InsetRect(8, 0);
	}
	bitmap.DrawText(pKeyboardWindow->m_font, rect, (const char *) m_text, CascadeColor(255, 255, 255, TRANSPARENCY), flags);
	if (IsActiveFocus()) Draw3DRect(bitmap, GetRectAbsolute());
}

///////////////////////////////////////////////////////////////////////////////
// class KeyboardWindow

KeyboardWindow::KeyboardWindow(u32 flags)
:	m_flags(flags)
,	m_bUseBackgroundBitmap(false)
,	m_highlightX(0)
,	m_highlightY(0)
,	m_lettersPerRow(10)
,	m_nRows(0)
,	m_callback(NULL)
,	m_nLetterWindows(0)
,	m_letters(NULL)
,	m_pActiveLetter(NULL)
,	m_bShift(false)
,	m_nButtonsInited(0)
{
    m_font.SetAttributes(CascadeFont::Attributes("univercb", CascadeFont::kNormal, 22));
	
	if (m_flags & KEYBOARD_NUMBERS)
	{
		if (m_flags & KEYBOARD_SHIFT_NUMBERS)
		{
			m_layoutstr       += "1234567890";
			m_layoutstr_shift += "!@#$%^&*()";
		}
		else
		{
			m_layoutstr_shift += "1234567890";
			m_layoutstr       += "1234567890";
		}
		m_nRows++;
	}

	if (m_flags & KEYBOARD_QWERTY)
	{
		m_layoutstr_shift += "QWERTYUIOPASDFGHJKL ZXCVBNM";
		m_layoutstr       += "qwertyuiopasdfghjkl zxcvbnm";
		m_nRows += 3;
	}
	else
	{
		m_layoutstr_shift += "ABCDEFGHIJKLMNOPQRSTUVWXYZ<>?:\"{}|_ +~";
		m_layoutstr       += "abcdefghijklmnopqrstuvwxyz,./;'[]\\- =`";
		m_nRows += 4;
	}

	// if no shifting, default to caps
	if ((m_flags & (KEYBOARD_SHIFT | KEYBOARD_SHIFT_NUMBERS)) == 0)
	{
		m_layoutstr = m_layoutstr_shift;
	}

	u32 numBigButtons = 0;
	if (m_flags & (KEYBOARD_SHIFT | KEYBOARD_SHIFT_NUMBERS)) numBigButtons++;
	if (m_flags & KEYBOARD_SPACE) numBigButtons++;
	if (m_flags & KEYBOARD_BACK) numBigButtons++;
	if (m_flags & KEYBOARD_DONE) numBigButtons++;
	if (numBigButtons > 1) m_nRows++;
}

KeyboardWindow::~KeyboardWindow()
{
	if (m_letters) delete [] m_letters;
}

void
KeyboardWindow::SetBackgroundBitmapFilename(CascadeString & path)
{
	CascadeFile file(path);
	m_bitmapBackground.Destroy();
	m_bUseBackgroundBitmap = m_bitmapBackground.CreateFromBitmapFile(file);
}

void
KeyboardWindow::RegisterOnEntryCallback(OnEntryProc callback, void * pClientData)
{
	m_callback = callback;
	m_pClientData = pClientData;
}

CascadeDims
KeyboardWindow::GetDims()
{
	if (m_bUseBackgroundBitmap) return m_bitmapBackground.GetDims();
	else return CascadeDims(kDefaultWidthPerChar * m_lettersPerRow, kDefaultHeightPerChar * m_nRows);
}

void
KeyboardWindow::Create()
{
	const CascadeRect myrect = GetRectAbsolute();
	m_stdButtonWidth = myrect.w / m_lettersPerRow;
	m_stdButtonHeight = myrect.h / m_nRows;

	u32 i;
	m_nLetterWindows = 0;
	const u32 layoutstrLen = m_layoutstr.GetLength();
	const char * pLayoutStr = (const char *) m_layoutstr;

	// count how many buttons we'll have
	for (i = 0; i < layoutstrLen; i++) if (pLayoutStr[i] != ' ') m_nLetterWindows++;
	if (m_flags & (KEYBOARD_SHIFT | KEYBOARD_SHIFT_NUMBERS)) m_nLetterWindows++;
	if (m_flags & KEYBOARD_SPACE) m_nLetterWindows++;
	if (m_flags & KEYBOARD_BACK) m_nLetterWindows++;
	if (m_flags & KEYBOARD_DONE) m_nLetterWindows++;

	// create and add them
	m_letters = new LilTextWindow[m_nLetterWindows];
	m_buttonRect = CascadeRect(0, 0, m_stdButtonWidth, m_stdButtonHeight);
	char buf[2] = { 'a', '\0' };

	for (i = 0; i < layoutstrLen; i++)
	{
		buf[0] = pLayoutStr[i];
		AddButton(buf);
	}
	if (m_flags & (KEYBOARD_SHIFT | KEYBOARD_SHIFT_NUMBERS)) AddButton("Shift");
	if (m_flags & KEYBOARD_SPACE) AddButton("Space");
	if (m_flags & KEYBOARD_BACK) AddButton("Back");
	if (m_flags & KEYBOARD_DONE) AddButton("Done");

	m_pActiveLetter = &m_letters[0];
}

bool
KeyboardWindow::AddButton(const CascadeString & text)
{
	if (m_nButtonsInited == m_nLetterWindows) return false;
	if (m_letters == NULL) return false;

	if (text != " ")
	{
		if (text.GetLength() > 1) m_buttonRect.w = 3 * m_stdButtonWidth;
		else m_buttonRect.w = m_stdButtonWidth;
		m_letters[m_nButtonsInited].SetText(text);
		m_letters[m_nButtonsInited].SetRectRelative(m_buttonRect);
		AddChild(&m_letters[m_nButtonsInited]);
		m_nButtonsInited++;
	}

	m_buttonRect.x += m_buttonRect.w;
	if (m_buttonRect.x >= GetRectRelative().w)
	{
		m_buttonRect.x = 0;
		m_buttonRect.y += m_buttonRect.h;
	}
	return true;
}

void
KeyboardWindow::ToggleShift()
{
	m_bShift = m_bShift ? false : true;
	SetButtonText();
}

void
KeyboardWindow::SetButtonText()
{
	if (m_letters)
	{
		const u32 layoutstrLen = m_layoutstr.GetLength();
		const char * pLayoutStr = (const char *) (m_bShift ? m_layoutstr_shift : m_layoutstr);
		char buf[2] = { 'a', '\0' };

		u32 i;
		u32 index = 0;
		for (i = 0; i < layoutstrLen; i++)
		{
			if (pLayoutStr[i] != ' ')
			{
				buf[0] = pLayoutStr[i];
				m_letters[index].SetText(buf);
				index++;
			}
		}
	}
}

void
KeyboardWindow::Activate()
{
	if (m_pActiveLetter) m_pActiveLetter->SetFocus();
}

void
KeyboardWindow::PaintBackground(CascadeBitmap & bitmap)
{
	CascadeRect rect = GetRectAbsolute();
	CascadeRect clipRect = bitmap.GetClipRect();
	if (rect.PointInRect(clipRect.TopLeft()) && rect.PointInRect(clipRect.BottomRight()))
	{
		// if clipRect is a sub-rect of our rect, then only render it
		rect = clipRect;
	}

	bitmap.FillRect(rect, CascadeColor(100, 100, 200, TRANSPARENCY));
}

void
KeyboardWindow::OnPaint(CascadeBitmap & bitmap)
{
	PaintBackground(bitmap);
}

bool
KeyboardWindow::OnKeyDown(u32 nKey)
{
	if (nKey == CK_SELECT)
	{
		if (m_pActiveLetter)
		{
			CascadeString text = m_pActiveLetter->GetText();
			//printf("got keypress \"%s\"\n", (const char *) text);
			if (text.GetLength() == 1)
			{
				if (m_callback) m_callback(text, m_pClientData);
			}
			else if (text == "Shift")
			{
				ToggleShift();
				Redraw();
			}
			else if (text == "Space")
			{
				if (m_callback) m_callback(CascadeString(" "), m_pClientData);
			}
			else if (text == "Back")
			{
				if (m_callback) m_callback(text, m_pClientData);
			}
			else if (text == "Done")
			{
				if (m_callback) m_callback(text, m_pClientData);
			}
		}
		return true;
	} else if (CK_MENU == nKey) {
   if (m_callback) m_callback(CascadeString("Done"), m_pClientData);
   return true;
 }
	return false;
}

bool
KeyboardWindow::OnNavKey(u32 nKey, CascadeWindow * & pProposedNewFocusWnd)
{
	if (IsParentOf(pProposedNewFocusWnd)) m_pActiveLetter = (KeyboardWindow::LilTextWindow *) pProposedNewFocusWnd;
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// class BGBmpKeyboardWindow

BGBmpKeyboardWindow::BGBmpKeyboardWindow(CascadeBitmap & bitmap)
:	m_bitmap(bitmap)
{
}

BGBmpKeyboardWindow::~BGBmpKeyboardWindow()
{
}

void
BGBmpKeyboardWindow::PaintBackground(CascadeBitmap & bitmap)
{
	const CascadeRect myrectrel = GetRectRelative();
	CascadeRect rect = GetRectAbsolute();
	CascadeRect clipRect = bitmap.GetClipRect();
	if (rect.PointInRect(clipRect.TopLeft()) && rect.PointInRect(clipRect.BottomRight()))
	{
		// if clipRect is a sub-rect of our rect, then only render it
		rect = clipRect;
	}

	// Assume bitmap is a background bitmap that covers our parent completely.
	// Therefore, our top left coord is our offset into the bitmap.
	CascadeRect srcRect = rect;
	srcRect.OffsetRect(myrectrel.x, myrectrel.y);
	//bitmap.Blit(rect.TopLeft(), m_bitmap, srcRect);

	bitmap.Blit(rect.TopLeft(), m_bitmap, rect);
}

