/**
 * VncSix main application and window.
 *
 * VncSix is a VNC viewer for the Roku PhotoBridge.
 *
 * Copyright (C) 2005, Andrew Oliver
 *
 * Portions based on videoscalerwindow.cpp, Copyright (c) 2004, Roku, LLC.
 *
 * Information about VncSix is available at: http://frequal.com/roku/VncSix/index.html
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cascade/Cascade.h>
#include <deschutes/libraries/hdmachinex225/VideoScaler.h>
#include <iostream>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>

#include "fbvnc.h"
#include "KeyboardWindow.h"
#include "ServerSelection.h"
#include "VncPrefs.h"
#include "VncSixWindow.h"

extern "C" {
  // From rfbproto.h
  #define CARD8 unsigned char
#define rfbButton1Mask 1
#define rfbButton2Mask 2
#define rfbButton3Mask 4
  #include "fbgl.h"
}

#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))

pthread_t threadVncComm;
int g_ciArgs;
char** g_ppszArgs;

extern "C" { 
  bool SendPointerEvent(int x, int y, int buttonMask);
  bool SendKeyEvent(CARD32 key, bool down);
}

const CARD32 KEY_NONE = 0;
const CARD32 PAGE_UP = 0xff55;
const CARD32 PAGE_DOWN = 0xff56;
const CARD32 ALT_L = 0xffe9;
const CARD32 LEFT_ARROW = 0xff51;
const CARD32 RIGHT_ARROW = 0xff53;
const CARD32 UP_ARROW = 0xff52;
const CARD32 DOWN_ARROW = 0xff54;
const CARD32 F4 = 0xffc1;
const CARD32 ESC = 0xff1b;
const CARD32 ENTER = 0xff0d;

typedef struct args_t {
  int argc;
  const char** argv;
  void* pvram;
  bool* pbStop;
} args_t;

extern "C" void* vncCommThreadWorker(void* pData) {
  struct args_t* pargs = (args_t*) pData;
  handleVncComm(pargs->pvram, pargs->pbStop);

  return NULL;
}

/**
 * The VncSix Cascade Application.
 *
 * Contains one VncSixWindow
 */
class VncSixApp : public CascadeApp
{
 public:
  virtual void OnAppInit();
  ServerSelection* pServerSelection;
 private:
  VncSixWindow m_window;
  int ciArgs;
  char** ppszArgs;
};

void VncSixApp::OnAppInit() {
  pServerSelection = new ServerSelection((CascadeWindow *)(&m_window), &m_window);
  pServerSelection->Setup();

  //m_window.m_colorFormat = VideoScaler::kColorFormatARGB8888_32; // default when no args
  m_window.Materialize();
}

VncSixApp app;

/**
 * Main routine for VncSix.
 *
 * Stores the program parameters and starts the Cascade application.
 */
int main(int argc, char ** argv) { 
  // Store paramters
  g_ciArgs = argc;
  g_ppszArgs = argv;

  // Start the application
  return app.Run(argc, (const char**) argv); 
}

/**
 * Print rectangle dimensions
 */
inline void DumpRect(const char * pName, const CascadeRect & rect) {
    printf("%s = (%d, %d, %d, %d)\n", pName, (int)rect.x, (int)rect.y, (int)rect.w, (int)rect.h);
}

/**
 * VncSix window constructor, sets up video scaler and starts VNC update thread.
 */
VncSixWindow::VncSixWindow() : 
  u32RepeatThreshold(500),
  u32MouseVelocityMax(48) {
  bKeyboardVisible = false;
  bKeyboardInitialized = false;
  m_colorFormat = VideoScaler::kColorFormatRGB565;
  m_nScalerBytesPerPixel = 0; // compute later
  m_pScalerVRAM = NULL;
  m_nScalerRowBytes = 0;
  m_bStopThread = false;
  u32MouseVelocity = 1;
  gettimeofday(&timeLastMouseMove, NULL);
  u32LastMouseXDelta = 0;
  u32LastMouseYDelta = 0;

  {  
    map<u32, Action*>* pmapRemoteActions1 = new map<u32, Action*>;
    map<u32, Action*>& mapRemoteActions = *pmapRemoteActions1;

    // Fill map with button/action pairs
    Action* paction = new Action();
    paction->key = PAGE_DOWN;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_ZOOM_OUT] = paction;

    paction = new Action();
    paction->key = PAGE_UP;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_ZOOM_IN] = paction;

    paction = new Action();
    paction->key = LEFT_ARROW;
    paction->keyModifier = ALT_L;
    mapRemoteActions[CK_PREVIOUS] = paction;

    paction = new Action();
    paction->key = F4;
    paction->keyModifier = ALT_L;
    mapRemoteActions[CK_PLAYPAUSE] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 1;
    mapRemoteActions[CK_SELECT] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 2;
    mapRemoteActions[CK_ROTATE] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 3;
    mapRemoteActions[CK_INFO] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 0;
    paction->mouseX = 1;
    paction->mouseY = 0;
    mapRemoteActions[CK_EAST] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 0;
    paction->mouseX = -1;
    paction->mouseY = 0;
    mapRemoteActions[CK_WEST] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 0;
    paction->mouseX = 0;
    paction->mouseY = 1;
    mapRemoteActions[CK_SOUTH] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 0;
    paction->mouseX = 0;
    paction->mouseY = -1;
    mapRemoteActions[CK_NORTH] = paction;

    mapKeymaps["Web"] = pmapRemoteActions1;

    // TODO: Allow change:
    pmapRemoteActions = pmapRemoteActions1;
  }

  {  
    map<u32, Action*>* pmapRemoteActions2 = new map<u32, Action*>;
    map<u32, Action*>& mapRemoteActions = *pmapRemoteActions2;

    // Fill map with button/action pairs
    Action* paction = new Action();
    paction->key = PAGE_DOWN;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_ZOOM_OUT] = paction;

    paction = new Action();
    paction->key = PAGE_UP;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_ZOOM_IN] = paction;

    paction = new Action();
    paction->key = ESC;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_PLAYPAUSE] = paction;

    paction = new Action();
    paction->key = ENTER;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_SELECT] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 2;
    mapRemoteActions[CK_ROTATE] = paction;

    paction = new Action();
    paction->key = KEY_NONE;
    paction->keyModifier = KEY_NONE;
    paction->mouseButton = 3;
    mapRemoteActions[CK_INFO] = paction;

    paction = new Action();
    paction->key = RIGHT_ARROW;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_EAST] = paction;

    paction = new Action();
    paction->key = LEFT_ARROW;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_WEST] = paction;

    paction = new Action();
    paction->key = DOWN_ARROW;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_SOUTH] = paction;

    paction = new Action();
    paction->key = UP_ARROW;
    paction->keyModifier = KEY_NONE;
    mapRemoteActions[CK_NORTH] = paction;

    mapKeymaps["MythTV"] = pmapRemoteActions2;
  }

  iterKeymaps = mapKeymaps.begin();
  iterKeymaps++;

  // Get the screen rectangle
  CascadeScreen screen;
  CascadeRect rectScreen(screen.GetRect());
  //CascadeRect rectScreen(0, 0, getWidth(), getHeight());

  // Set up scaling from VRAM buffer size to screen size
  u32 nIndex = 0;
  CascadeRect scaleRect;

  // figure out what output resolution [index] the screen is set to.
  screen.GetCurrentOutputResolution(nIndex);

  // get the current scale rect for that output resolution
  screen.GetOutputResolutionScaleRect(nIndex, scaleRect);
    
  // just for fun, print out the ScreenRect (CascadeWindow coordinate system)
  // and the scale rect (CascadeScreen output resolution coordinate system)
  DumpRect("ScreenRect", rectScreen);
  DumpRect(" ScaleRect", scaleRect);
  
  // do some floating point
  double scaleHorz = (double)scaleRect.w / (double)rectScreen.w;
  double scaleVert = (double)scaleRect.h / (double)rectScreen.h;
    
  // Make the window fill the entire screen
  SetRectAbsolute(rectScreen);          // set it as my CascadeWindow's rectangle

  // now we're ready to recast our CascadeWindow rectangle (rect) into a rectangle
  // of the same size and position in the VideoScaler (output scale rect) coordinate space.
  m_scalerOutputRect.x = (s32)((double)rectScreen.x * scaleHorz + (double)scaleRect.x + (double)0.5);
  m_scalerOutputRect.y = (s32)((double)rectScreen.y * scaleVert + (double)scaleRect.y + (double)0.5);
  m_scalerOutputRect.w = (u32)((double)rectScreen.w * scaleHorz + (double)0.5);
  m_scalerOutputRect.h = (u32)((double)rectScreen.h * scaleVert + (double)0.5);

  DumpRect("WindowRect", rectScreen);
  DumpRect("OutputRect", m_scalerOutputRect);

  // Keyboard setup
  CascadeRect rect(screen.GetRect());
  SetRectAbsolute(CascadeRect(0, 0, rect.w, rect.h));
  SetFocus();
  
  // Create keyboard window and register key handler
  u32 flags = 
    KEYBOARD_NUMBERS | 
    KEYBOARD_SHIFT_NUMBERS | 
    KEYBOARD_SHIFT | 
    KEYBOARD_BACK | 
    KEYBOARD_SPACE | 
    KEYBOARD_DONE;
  m_pKeyboard = new KeyboardWindow(flags);
  m_pKeyboard->RegisterOnEntryCallback(&keyHandler, this);    
  AddChild(m_pKeyboard);
}

/**
 * When this class is destroyed, release the allocated scaler resources.
 */
VncSixWindow::~VncSixWindow() {
  if (NULL != m_pScalerVRAM) {
    //pthread_join(threadVncComm, NULL);
    m_scaler.FreeScalerVRAM(m_pScalerVRAM);
    m_scaler.Close();
    m_pScalerVRAM = NULL;

    delete m_pKeyboard;
    m_pKeyboard = NULL;
  }
}

/**
 * Ask the user which server should be used via the GUI, allowing
 * entry of a new one.
 */
void VncSixWindow::promptForServer() {
  app.pServerSelection->Materialize();
  app.pServerSelection->RedrawScreen();
  
}

void VncSixWindow::hideKeyboard() {
  printf("Keyboard being hidden...");

  // Hide the window
  CascadeDims dims = m_pKeyboard->GetDims();                      
  CascadeRect keyboardRect(2000, 1000, dims.w, dims.h);
  m_pKeyboard->SetRectRelative(keyboardRect);
  //RemoveChild(m_pKeyboard);
  Redraw();
  SetFocus();
  
  bKeyboardVisible = false;
}

void VncSixWindow::nextKeymap() {
  if (iterKeymaps != mapKeymaps.end()) {
    iterKeymaps++;
  }

  if (iterKeymaps == mapKeymaps.end()) {
    iterKeymaps = mapKeymaps.begin();
  }

  pmapRemoteActions = iterKeymaps->second;
}

void VncSixWindow::setKeymap(std::string strKeymapName) {
  map<std::string, map<u32, Action*>*>::iterator iterKeymapsTemp = mapKeymaps.find(strKeymapName);
  if (iterKeymapsTemp != mapKeymaps.end()) {
    printf("Setting keymap to %s\n", strKeymapName.c_str());
    pmapRemoteActions = mapKeymaps[strKeymapName];
    iterKeymaps = iterKeymapsTemp;
  }
}

void VncSixWindow::setupVncWindow() {
  CascadeRect rect(GetRectAbsolute());
  // Take dimensions from the VNC server
  printf("dims: %d x %d\n", getWidth(), getHeight());
  CascadeDims dims(getWidth(), getHeight());
  u32MouseX = getWidth() / 2;
  u32MouseY = getHeight() / 2;
  SendPointerEvent(u32MouseX, u32MouseY, 0);
  
  // open and configure the video scaler
    
    // first figure out bytes/pixel that we're using
  switch (m_colorFormat) {
  case VideoScaler::kColorFormatRGB565:
    m_nScalerBytesPerPixel = 2;
    printf("\nUsing kColorFormatRGB565\n");
    break;
  case VideoScaler::kColorFormatARGB8888_32:
    m_nScalerBytesPerPixel = 4;
    printf("\nUsing kColorFormatARGB8888_32\n");
    break;
  default:
    break;
  }

  bool bSuccess = false;
  do {
    // make sure we're using a valid m_colorFormat
    if (0 == m_nScalerBytesPerPixel) break;

    // open up the scaler in that format
    if (! m_scaler.Open(m_colorFormat)) break;
        
    // set the dimensions of the source memory.  
    // We can choose any dimensions that we want - the video scaler will
    // scale this source memory to the m_scalerOutputRect we computed in
    // the constructor.  In our case, we'll use the dimensions of the
    // CascadeWindow so that we can stay living in the CascadeWindow coordinate
    // space.
    u32 nPixelPitch = 0;
    if (! m_scaler.SetSourceWindow(dims, nPixelPitch)) break;

    // ok, we told the scaler what we wanted (dims) and it gave us back
    // a line-width for memory allocation in terms of pixels (in the current color mode)
    // convert this to a byte pitch:
    m_nScalerRowBytes = nPixelPitch * m_nScalerBytesPerPixel;
    global_framebuffer.v_xsize = nPixelPitch;

    // now we know how much memory to allocate - allocate it
    if (NULL == (m_pScalerVRAM = m_scaler.AllocateScalerVRAM(dims.h * m_nScalerRowBytes))) break;
        
    // Setup scaler output rect to match display
    CascadeRect scaleRect;
    // figure out what output resolution [index] the screen is set to.
    CascadeScreen screen;
    u32 nIndex = 0;
    screen.GetCurrentOutputResolution(nIndex);
    // get the current scale rect for that output resolution
    screen.GetOutputResolutionScaleRect(nIndex, scaleRect);

    // Compute how to scale and position output rect to center and preserve aspect ratio
    // ...Figure out which dimension has to grow less, use that as the scaling factor
    // ...(Divide screen size by server size to find out how much to scale server to screen)
    double dScalingX = ((double) scaleRect.w) / ((double) getWidth());
    double dScalingY = ((double) scaleRect.h) / ((double) getHeight());
    double dScalingFactor = min(dScalingX, dScalingY);
    // ...Multiply both by scaling factor to get scaled width and height
    m_scalerOutputRect.w = (u32) (getWidth() * dScalingFactor);
    m_scalerOutputRect.h = (u32) (getHeight() * dScalingFactor);
    // ...Subtract scaled width from scaleRect.w, /2, add scaleRect.x to get new centered x
    m_scalerOutputRect.x = (scaleRect.w - m_scalerOutputRect.w) / 2 + scaleRect.x;
    // ...Repeat for y
    m_scalerOutputRect.y = (scaleRect.h - m_scalerOutputRect.h) / 2 + scaleRect.y;

    // HACK: Only 1024 wide seems to scale properly now, scale others 1:1
    /*
      if (getWidth() < 1024) {
      m_scalerOutputRect.w = 1024;
      m_scalerOutputRect.h = getHeight();
      m_scalerOutputRect.x = 0;
      m_scalerOutputRect.y = 0;
      }
    */

    // tell the scaler about the memory we just allocated
    if (! m_scaler.SetSourceVRAM(m_pScalerVRAM)) break;

    // tell the scaler about the output rectangle we computed in the constructor
    if (! m_scaler.SetDestWindow(m_scalerOutputRect, false)) break;

    // crop/zoom the scaler output 1:1 - this is a tricky one
    // what it does is actually take what would be imaged to the output
    // rectangle and crop that result (after scaler pass 1) to the crop
    // rectangle we're setting with SetCropWindow().  It still displays
    // the output in the rect set with SetDestWindow(), so the result is
    // actually zoomed in or zoomed out data.
    if (! m_scaler.SetCropWindow(m_scalerOutputRect)) break;

    // Fill the screen with gray
    memset(m_pScalerVRAM, 0x88, dims.h * m_nScalerRowBytes);
        
    // scaler is ready to go - let's display it on the screen
    if (! m_scaler.Show(VideoScaler::kBlendModeAdd)) break; // using scaler for graphics, not video
        
    printf("Setup video scaler OK\n");
    bSuccess = true;
  } while (false);

  if (bSuccess) {
    args_t* pargs = new args_t;
    pargs->argc = g_ciArgs;
    pargs->argv = (const char**) g_ppszArgs;
    // Wait for the video RAM to become setup, should be unecessary now
    for (int i = 0; i < 15; i++) {
      if (NULL != m_pScalerVRAM) {
        break;
      }
      sleep(1);
    }
    if (NULL == m_pScalerVRAM) {
      printf("app.cpp, thread args filling: pVideoRam is NULL!");
    }
    pargs->pvram = m_pScalerVRAM;
    pargs->pbStop = &(m_bStopThread);
      
    // Start VNC communications thread
    printf("OnAppInit: about to start thread\n");
    pthread_create(&threadVncComm,
                   NULL,
                   &vncCommThreadWorker,
                   (void *) pargs);
    
    printf("Setting focus\n");
    SetFocus();
  } else {
    if (m_scaler.IsOpen()) m_scaler.Close();
    if (NULL != m_pScalerVRAM) m_scaler.FreeScalerVRAM(m_pScalerVRAM);
    m_pScalerVRAM = NULL;
    OnKeyDown(CK_EXIT);
  }
}

void VncSixWindow::OnMaterialize() {
  // Setup keymap if specified
  for (int i = 0; i < g_ciArgs; i++) {
    cout << " Checking arg: "  << i << endl;
    if (!strcmp("-m", g_ppszArgs[i])) {
      cout << " -m found " << endl;
      if ((i+1) <= g_ciArgs) {
        setKeymap(std::string(g_ppszArgs[i+1]));
      }
      break;
    }
  }

  // Make connection to the server specified on the command line or interactively
  int iStatus = connectToVncServer(g_ciArgs, g_ppszArgs);
  
  if (iStatus) {
    cout << "Couldn't connect, asking for user input" << endl;
    promptForServer();
  } else {
    cout << "Connected, showing the window" << endl;
    setupVncWindow();
  }
}

/** 
 * Append text to internal list and force repaint
 */
void VncSixWindow::OnNewText(CascadeString str) {
  if (str != "Done") {
    char cPressed = str[0];
    // TODO: Send key to server
    SendKeyEvent(cPressed, 1);
    SendKeyEvent(cPressed, 0);
  } else {
    hideKeyboard();
  }
}

void VncSixWindow::OnVanish() {
  // Signal to the VNC message handler thread to terminate
  m_bStopThread = true;

  // End the current application (VncSixApp)
  CascadeApp::GetApp()->Terminate(0);
}

/**
 * Fill this window with a fully transparent color to let the video
 * pane show through.  The important part is the alpha value of zero,
 * meaning let the video pass through 100%.
 */
void VncSixWindow::OnPaint(CascadeBitmap& bitmap) {
  bitmap.FillRect(GetRectAbsolute(), CascadeColor(0, 0, 0, 0));
}

void VncSixWindow::handleRemoteButton(u32 button) {
  if (pmapRemoteActions->find(button) != pmapRemoteActions->end()) {
    Action* paction = (*pmapRemoteActions)[button];

    if (paction) {
      if (paction->key) {
        sendKey(paction->key, paction->keyModifier);
      } else if (paction->mouseButton) {
        sendMouseButtonClick(paction->mouseButton);
      } else {
        moveMouse(paction->mouseX, paction->mouseY);
      }
    }
  }
}

void VncSixWindow::sendKey(CARD32 key, CARD32 keyModifier = 0) {
  if (keyModifier) {
    SendKeyEvent(keyModifier, 1);
  }
  SendKeyEvent(key, 1);
  SendKeyEvent(key, 0);
  if (keyModifier) {
    SendKeyEvent(keyModifier, 0);
  }
}

/*
void VncSixWindow::sendPageUp() {
  sendKey(PAGE_UP);
}

void VncSixWindow::sendPageDown() {
  sendKey(PAGE_DOWN);
}

void VncSixWindow::sendAltLeftArrow() {
  sendKey(LEFT_ARROW, ALT_L);
}

void VncSixWindow::sendAltF4() {
  sendKey(F4, ALT_L);
}
*/

void VncSixWindow::sendMouseButtonClick(int mouseButton) {
  SendPointerEvent(u32MouseX, u32MouseY, 1 << (mouseButton - 1));
  SendPointerEvent(u32MouseX, u32MouseY, 0);
}

/*
void VncSixWindow::sendLeftMouseButtonClick() {
  SendPointerEvent(u32MouseX, u32MouseY, rfbButton1Mask);
  SendPointerEvent(u32MouseX, u32MouseY, 0);
}

void VncSixWindow::sendMiddleMouseButtonClick() {
  SendPointerEvent(u32MouseX, u32MouseY, rfbButton2Mask);
  SendPointerEvent(u32MouseX, u32MouseY, 0);
}

void VncSixWindow::sendRightMouseButtonClick() {
  SendPointerEvent(u32MouseX, u32MouseY, rfbButton3Mask);
  SendPointerEvent(u32MouseX, u32MouseY, 0);
}
*/

void VncSixWindow::moveMouse(int iXDelta, int iYDelta) {
  // Determine if button is being held down
  timeval timeNow;
  gettimeofday(&timeNow, NULL);
  long lTimeDelta = abs((timeNow.tv_sec - timeLastMouseMove.tv_sec) * 1000 + 
                        (timeNow.tv_usec - timeLastMouseMove.tv_usec) / 1000);
  if (lTimeDelta < u32RepeatThreshold) {
    // Repeating, increase velocity
    u32MouseVelocity += 2;
    if (u32MouseVelocity > u32MouseVelocityMax) {
      u32MouseVelocity = u32MouseVelocityMax;
    }
  } else {
    // No repeat, set velocity to base
    u32MouseVelocity = 1;
  }    
  printf("timeNow=%d/%d, timeLast=%d/%d, timeDelta=%ld, velocity=%d\n", 
         timeNow.tv_sec, timeNow.tv_usec, 
         timeLastMouseMove.tv_sec, timeLastMouseMove.tv_usec, 
         lTimeDelta, u32MouseVelocity);
  gettimeofday(&timeLastMouseMove, NULL);

  u32MouseX += iXDelta * u32MouseVelocity;
  u32MouseY += iYDelta * u32MouseVelocity;

  // Make sure we don't move offscreen
  if (u32MouseX >= getWidth()) {
    if (iXDelta < 0) {
      u32MouseX = 0;
    } else {
      u32MouseX = getWidth() - 1;
    }
  }
  if (u32MouseY >= getHeight()) {
    if (iYDelta < 0) {
      u32MouseY = 0;
    } else {
      u32MouseY = getHeight() - 1;
    }
  }

  SendPointerEvent(u32MouseX, u32MouseY, 0);
}

/*
void VncSixWindow::moveNorth() {
  moveMouse(0, -1);
}

void VncSixWindow::moveSouth() {
  moveMouse(0, 1);
}

void VncSixWindow::moveWest() {
  moveMouse(-1, 0);
}

void VncSixWindow::moveEast() {
  moveMouse(1, 0);
}
*/

/**
 * Handle button presses, exit key terminates the program.
 */
bool VncSixWindow::OnKeyDown(u32 u32Key) {
  switch (u32Key) {
    case CK_EXIT: {
      CascadeApp::GetApp()->Terminate(0);
      //Vanish();
    } break;
    case CK_MENU: {
      if (bKeyboardVisible) {
        hideKeyboard();
      } else {
        // Show the keyboard
        CascadeDims dims = m_pKeyboard->GetDims();                      
        CascadeRect keyboardRect(750, 24, dims.w, dims.h);
        m_pKeyboard->SetRectRelative(keyboardRect);
        if (!bKeyboardInitialized) {
          m_pKeyboard->Create();
          bKeyboardInitialized = true;
        }
        m_pKeyboard->Activate();
        Redraw();

        bKeyboardVisible = true;
      }

      //sendRightMouseButtonClick();
    } break;
    case CK_NEXT: {
      nextKeymap();
    } break;
    case CK_INFO:
    case CK_ROTATE:
    case CK_SELECT:
    case CK_ZOOM_IN:
    case CK_ZOOM_OUT:
    case CK_PREVIOUS:
    case CK_PLAYPAUSE: {
      handleRemoteButton(u32Key);
    } break;
  }
  
  return true;
}

bool VncSixWindow::OnNavKey(u32 nKey, CascadeWindow *&pProposedNewFocusWnd) {
  if (bKeyboardVisible) {
    // Keyboard shown
    if (m_pKeyboard->IsParentOf(CascadeWindow::GetFocus())) {
      if (m_pKeyboard->IsParentOf(pProposedNewFocusWnd)) {
        return false;
      } else {
        // don't let the user arrow off the keyboard
        pProposedNewFocusWnd =
          CascadeWindow::GetFocus();
        return true;
      }
    }
  } else {
    // No keyboard shown, move mouse
    switch (nKey) {
      case CK_EAST:
      case CK_WEST:
      case CK_NORTH:
      case CK_SOUTH: {
        handleRemoteButton(nKey);
      } break;
      default: {
        //cout << "Key pressed.  key code=" << nKey << endl;
      } break;
    }
  }
  
  return true; // Report that we handled the keypress
}

void keyHandler(const CascadeString& str, void* pData) {
  VncSixWindow* pvnc = (VncSixWindow*) pData;

  pvnc->OnNewText(str);
}
