/*
 *  Copyright (C) 2005 Andrew Oliver
 *  Previous version Copyright (C) Klaus Weidner
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <sys/shm.h>

#include "vncviewer.h"
#include "fbgl.h"
#include "list.h"

fbvnc_framebuffer_t global_framebuffer;

static bool quit_requested = 0;

bool landscape_is_native = 0;

void fbvnc_close(void);

void cleanup_and_exit(char *msg, int ret) {
  // TODO: Show error message on screen instead
  //if (ret==EXIT_SYSERROR) {
  //if (msg) perror(msg); else perror("error");
  //} else {
    if(msg) fprintf(stderr, "%s\n", msg);
    //}
  fbvnc_close();
  //exit(ret);
}

void signal_quit(int sig) {
  quit_requested = 1;
}

void* xmalloc(size_t s) {
  void *r;
  r=malloc(s);
  if (!r) cleanup_and_exit("out of memory", 1);
  return r;
}

void get_fbinfo() {
  myFormat.bitsPerPixel = 8*sizeof(Pixel);
  myFormat.depth = 16;
  myFormat.bigEndian = 0;
  myFormat.trueColour = 1;
  myFormat.redShift = 11;
  myFormat.redMax = 31;
  myFormat.greenShift = 5;
  myFormat.greenMax = 63;
  myFormat.blueShift = 0;
  myFormat.blueMax = 31;

  // Make the same size as the VNC server
  //global_framebuffer.p_xsize = si.framebufferWidth;
  //global_framebuffer.p_ysize = si.framebufferHeight;
}

void fbvnc_init(void* pVideoRam) {
  int v_xsize = si.framebufferWidth;
  int v_ysize = si.framebufferHeight;

  global_framebuffer.p_framebuf = 0;
  global_framebuffer.num_read_fds = 0;
  global_framebuffer.num_write_fds = 0;

  global_framebuffer.ts_x = 0;
  global_framebuffer.ts_y = 0;

  //global_framebuffer.v_xsize = v_xsize;
  global_framebuffer.v_ysize = v_ysize;
  global_framebuffer.v_x0 = 0;
  global_framebuffer.v_y0 = 0;
  global_framebuffer.v_scale = 1;
  global_framebuffer.v_bpp = sizeof(Pixel);

  global_framebuffer.v_buf = pVideoRam;

  signal(SIGINT, signal_quit);
  signal(SIGTERM, signal_quit);
  signal(SIGQUIT, signal_quit);

  /* start in portrait mode by default ? */
  global_framebuffer.p_landscape = landscape_is_native;

  get_fbinfo();
}

void fbvnc_close() {
  if (global_framebuffer.p_framebuf) {
    global_framebuffer.p_framebuf = 0;
  }
}

#define SQR(x) ((x)*(x))

struct sched {
  int tv_sec;
  int tv_usec;
  int event;
};

int cmp_sc(struct sched *A, struct sched *B) {
  if (A->tv_sec < B->tv_sec) return -1;
  if (A->tv_sec > B->tv_sec) return  1;
  if (A->tv_usec < B->tv_usec) return -1;
  if (A->tv_usec > B->tv_usec) return  1;
  return 0;
}

void schedule_add(List *s, int ms_delta, int event) {
  struct timeval tv;
  struct sched sc;
  List *p;

  gettimeofday(&tv, 0);
  sc.tv_sec = tv.tv_sec;
  sc.tv_usec = tv.tv_usec;
  sc.event = event;

  sc.tv_usec += (ms_delta%1000)*1000;
  while (sc.tv_usec > 1000000) {
    sc.tv_usec -= 1000000;
    ++ sc.tv_sec;
  }
  sc.tv_sec += ms_delta / 1000;

  /* insert into sorted list */
  for (p=s; p->next; p=p->next) {
    if (cmp_sc(&sc, p->next->val) < 0) {
      list_insert_copy(p, &sc, sizeof sc);
      break;
    }
  }
  if (!p->next) {
    list_append_copy(s, &sc, sizeof sc);
  }
}

void schedule_delete(List *s, int event) {
  List *p;

  for (p=s->next; p; p=p->next) {
    if (p->val) {
      struct sched *n = p->val;
      if (n->event == event) break;
    }
  }

  if (p) list_remove(s, p);
}

int schedule_get_next(List *s, struct timeval *tv) {
  List *p = s->next;
  struct sched *nv;

  if (!p) return 0;

  nv = (struct sched *)p->val;
  gettimeofday(tv, 0);
  tv->tv_sec = nv->tv_sec - tv->tv_sec;
  tv->tv_usec = nv->tv_usec - tv->tv_usec;

  while (tv->tv_usec < 0) {
    tv->tv_usec += 1000000;
    -- tv->tv_sec;
  }

  while (tv->tv_usec > 1000000) {
    tv->tv_usec -= 1000000;
    ++ tv->tv_sec;
  }

  if (tv->tv_sec < 0) {
    tv->tv_sec = 0;
    tv->tv_usec = 0;
  }
  return nv->event;
}

void fd_copy(fd_set *out, fd_set *in, int n) {
  int i;
  FD_ZERO(out);
	
  for (i=0; i<n; i++) {
    if (FD_ISSET(i, in)) FD_SET(i, out);
  }
}

#define FD_SET_u(fd, set) do {			\
		FD_SET((fd), (set));		\
		if ((fd) > max) max=(fd);	\
	} while(0)

#define RetEvent(e) do { ev->evtype = (e); return (e); } while(0)

void scaledPointerEvent(int xp, int yp, int button) {
  IMPORT_FRAMEBUFFER_VARS

  /* fprintf(stderr, "mouse click, button=%d\n", button); */
	 if (p_landscape) {
    SendPointerEvent(xp*v_scale + v_x0, yp*v_scale + v_y0, button);
  } else {
    SendPointerEvent((yp          )*v_scale + v_x0,
                     (p_xsize-xp-1)*v_scale + v_y0,
                     button);
  }
}

bool img_saved = 0;
List *sched;

#define FDNUM_VNC 0
#define FDNUM_PNM 1

void save_viewport(struct viewport *dst) {
  IMPORT_FRAMEBUFFER_VARS

 	dst->v_xsize = v_xsize;
  dst->v_ysize = v_ysize;
  dst->v_buf = v_buf;
  dst->v_scale = v_scale;
  dst->p_landscape = p_landscape;
}

void activate_viewport(struct viewport *src) {
  global_framebuffer.v_xsize = src->v_xsize;
  global_framebuffer.v_ysize = src->v_ysize;
  global_framebuffer.v_buf = src->v_buf;
  global_framebuffer.v_scale = src->v_scale;
  global_framebuffer.p_landscape = src->p_landscape;
}

enum fbvnc_event fbvnc_get_event (fbvnc_event_t *ev, List *sched) {
	fd_set rfds, wfds, save_rfds;
	int i, ret;
	int max;
	struct timeval timeout;
	struct timeval tzero = {0, 0};
	bool got_data_immediately;
#ifdef INPUT_PS2MOUSE
	extern int msefd; /* set in ofbis library */
#endif
	IMPORT_FRAMEBUFFER_VARS

	if (sched) {
		int s_event;

		s_event = schedule_get_next(sched, &timeout);

		if (s_event) {
			ev->evtype = s_event;
		} else {
			ev->evtype = FBVNC_EVENT_TIMEOUT;
		}
	} else {
		ev->evtype = FBVNC_EVENT_TIMEOUT;
	}
	if (ev->evtype == FBVNC_EVENT_TIMEOUT) {
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
	}
	
	max = 0;

	FD_ZERO(&rfds);
	FD_ZERO(&wfds);

#ifdef INPUT_TS
	FD_SET_u(ts_fd, &rfds);
#endif
#ifdef INPUT_PS2MOUSE
	FD_SET_u(msefd, &rfds);
#endif
	FD_SET_u(kb_fd, &rfds);

	/* quick check on input devices without waiting */
	fd_copy(&save_rfds, &rfds, max+1); /* save for later */
	ret = select(max+1, &rfds, 0, 0, &tzero);

	if (ret) {
		got_data_immediately = 1;
	} else {
		got_data_immediately = 0;

		/* try again on all fds, wait this time */
		fd_copy(&rfds, &save_rfds, max+1); /* restore */

		for(i=0; i<num_read_fds; i++) {
			if (read_fd[i] < 0) continue;
			FD_SET_u(read_fd[i], &rfds);
		}

		for(i=0; i<num_write_fds; i++) {
			if (write_fd[i] < 0) continue;
			FD_SET_u(write_fd[i], &wfds);
		}
		ret = select(max+1, &rfds, &wfds, 0, &timeout);
	}

	if (!ret) {
		if (ev->evtype != FBVNC_EVENT_TIMEOUT) {
			schedule_delete(sched, ev->evtype);
		}
		return ev->evtype;
	}

	if (quit_requested) {
		cleanup_and_exit("Interrupt.", EXIT_OK);
	}

	ev->x = ts_x;
	ev->y = ts_y;
	ev->dx = 0;
	ev->dy = 0;

	if (FD_ISSET(kb_fd, &rfds)) {
		unsigned char k;
		int r;
		r=read(kb_fd, &k, sizeof k);
		if (r!=sizeof k) cleanup_and_exit("read kb", EXIT_SYSERROR);

		if (debug) {
			/* debug keyboard */
			fprintf(stderr, "key=%d (0x%02x)\n", k, k);
		}

		ev->key = k&0x7f;
		RetEvent((k&0x80) ? FBVNC_EVENT_BTN_UP : FBVNC_EVENT_BTN_DOWN);
	}
	for (i=0; i<num_read_fds; i++) {
		int fd = read_fd[i];
		if (fd<0) continue;
		if (FD_ISSET(fd, &rfds)) {
			ev->fd = fd;
			RetEvent(FBVNC_EVENT_DATA_READABLE);
		}
	}
	for (i=0; i<num_write_fds; i++) {
		int fd = write_fd[i];
		if (fd<0) continue;
		if (FD_ISSET(fd, &wfds)) {
			ev->fd = fd;
			RetEvent(FBVNC_EVENT_DATA_WRITABLE);
		}
	}
	cleanup_and_exit("unhandled event", EXIT_ERROR);
	return 0;
}

static struct viewport v_img;

void processArgs(int argc, char **argv);

int connectToVncServerRaw() {
  int iResult = 0;

  printf("Connecting to server %s...  ", hostname);
  if (ConnectToRFBServer(hostname, port)) {
    printf("Connected to server %s.\n", hostname);

    printf("Initializing connection...  ");
    if (InitialiseRFBConnection(pszServer, rfbsock)) {
      printf("Initialized connection.\n");
    } else {
      // Couldn't initialize
      printf("Couldn't initialize\n");
      iResult = 2;
    }
  } else {
    // Couldn't connect
    printf("Couldn't connect\n");
    iResult = 1;
  }

  return iResult;
}

int connectToVncServer(int argc, char **argv) {
  processArgs(argc, argv);
  
  return connectToVncServerRaw();
}

unsigned short getWidth() {
  return si.framebufferWidth;
}

unsigned short getHeight() {
  return si.framebufferHeight;
}

int handleVncComm(void* pVideoRam, bool* pbStop) {
  int readfd[2]; /* pnmfd and/or rfbsock */

  sched = list_new();

  if (NULL == pVideoRam) {
    printf("fbvnc.c: pVideoRam is NULL!");
    return 1;
  }
  fbvnc_init(pVideoRam);

  if (!SetFormatAndEncodings()) 
    cleanup_and_exit("encodings", EXIT_ERROR);

  if (!SendFramebufferUpdateRequest(0, 0, si.framebufferWidth,
                                    si.framebufferHeight, False)) 
    cleanup_and_exit("update request", EXIT_ERROR);

  global_framebuffer.num_read_fds = 2;
  global_framebuffer.num_write_fds = 0;
  global_framebuffer.read_fd = readfd;
  readfd[FDNUM_VNC] = rfbsock;
  readfd[FDNUM_PNM]=-1;

  schedule_add(sched, 1000, FBVNC_EVENT_TICK_SECOND);

  while(1) {
    fd_set rfds;
    struct timeval tv;
    int retval;

    printf("In VNC handler loop\n");
   
    SendIncrementalFramebufferUpdateRequest();

    /*
      //printf("Calling SendFramebufferUpdateRequest");
      //SendFramebufferUpdateRequest(0, 0, si.framebufferWidth, si.framebufferHeight, False);
    */

    /* Watch stdin (fd 0) to see when it has input. */
    FD_ZERO(&rfds);
    FD_SET(rfbsock, &rfds);
    /* Wait up to one second. */
    tv.tv_sec = 0;
    tv.tv_usec = 50;

    retval = select(1, &rfds, NULL, NULL, &tv);
    /* Don't rely on the value of tv now! */

    //if (retval) {
    if (1) {
      printf("Calling HandleRFBServerMessage...  ");
      if (!HandleRFBServerMessage())
        cleanup_and_exit("rfb msg", EXIT_ERROR);
      printf("HandleRFBServerMessage returned.\n");
    }

    if (*pbStop) {
      break;
    }
  }
 
  fbvnc_close();
 
  return 0;
}
